import os
from bs4 import BeautifulSoup

def GetConfig(configName):
    if 'PROJECTFOLDER' in os.environ:
        configPath = os.environ.get('PROJECTFOLDER')
        if configPath[-1] == '/':
            configPath = configPath[:-1]
    else:
        raise EnvironmentError('No PROJECTFOLDER environment variable set.  Can not find Config.xml')

    if configName.lower() == 'projectfolder':
        return configPath

    if os.path.exists(configPath):
        config = open(configPath + '/Config.xml', 'r')
    else:
        config = open(configPath + '/Config.xml', 'w')

    soup = BeautifulSoup(config, 'xml')
    node = soup.find(configName)

    if node==None:
        raise ValueError("Could not find value in Config.XML for [%s]" % configName)
    return node.text



import datetime
import requests
import os
import base64
import sys

from hotdog.FilePath import get_full_path
from hotdog.Config import GetConfig
import time
MustardURL = GetConfig('MUSTARD_URL')
MustardKey = GetConfig('MUSTARD_KEY')

PROJECTFOLDER = GetConfig('ProjectFolder')

def takeScreenshot(driver, imageName):
    worked = True
    try: driver.save_screenshot(imageName)
    except:
        try: driver.get_screenshot_as_file(imageName)
        except: worked = False
    return worked

def UploadToMustard(test, status, error_message=None, stacktrace=None):

    if test.options['mustard']:
        try:

            imageName = PROJECTFOLDER + str(int(round(time.time() * 1000)))+'.png'
            if takeScreenshot(test.driver, imageName):
                with open(imageName, "rb") as image_file:
                  files = base64.b64encode(image_file.read())
                files = "data:image/png;base64," + files.decode()
            else:
                files = None
        except:
            files = None


        if test.testcase_id:
            test_identifier = test.testcase_id
        else:
            test_identifier = test._testMethodName

        platform = test.desired_caps['platformName'] if 'platformName' in test.desired_caps else test.desired_caps['platform']
        payload = {"result":{ "project_id": MustardKey,
                              "result_type": "automated",
                               "environment_id": getDeviceID(test),
                               "testcase_id": test_identifier,
                               "status": status,
                               "comment": error_message,
                               "screenshot": files,
                               "stacktrace": stacktrace,
                               "display_name": test.options["manufacturer"] + " " +test.options["model"] + " " + test.options["osv"],
                               "link": test.resultLink,
                               "step_log": test.driver.step_log.to_json()
                      }
                 }
        if files:
            Upload(payload)
            # Upload(payload, files)
        else:
            Upload(payload)

        try:
            os.remove(imageName)
        except:
            pass

# def UploadScreenshot(self, test, name=None):
#     imageName = PROJECTFOLDER + str(int(round(time.time() * 1000)))+'.png'
#     if test.options['mustard']:
#         if takeScreenshot(test.driver, imageName):
#             files = {'screenshot': open(imageName, 'rb')}
#
#             payload = {'project_id': MustardKey,
#                    'result_type': 'screenshot',
#                    'device_id': getDeviceID(test),
#                    'test_name': name if name else test.__class__.__name__,
#                    }
#
#             Upload(payload, files)
#
#         try:
#             os.remove(imageName)
#         except:
#             pass
#
#
#
# def UploadPerformance(device, name, time):
#
#     payload = {'project_id': MustardKey,
#                'result_type': 'performance',
#                'time': time,
#                'device_id': device,
#                'test_name': name,
#                }
#
#     Upload(payload)


def Upload(payload, files=None):
    caughtException = False
    try:
        r = requests.post(MustardURL, json=payload, files=files)

    except:
        exceptionMessage = str(sys.exc_info()[1])
        caughtException = True
        print(exceptionMessage)

    if  caughtException or r.status_code != 200:
        bl = PROJECTFOLDER + '/MustardFailSafe.txt'

        with open(bl, 'a') as backlog:
            backlog.write(str(datetime.datetime.now()) + '\n')
            backlog.write(str(payload))
            if 'exceptionMessage' in locals():
                backlog.write(exceptionMessage)
            backlog.write('\n')
            backlog.write('\n')
        print('Failed to upload results to mustard.  Saved to MustardFailSafe.txt')

def getDeviceID(test):
    try:
        if test.driver and 'udid' in test.driver.desired_capabilities:
            return test.driver.desired_capabilities['udid']
        else:
            return test.options['deviceName']
    except:
        return 'ERROR'


from selenium.common.exceptions import StaleElementReferenceException

class wait_for_present(object):
    def __init__(self, element):
        self.element = element

    def __call__(self, driver):
        try:
            self.element.load()
            if self.element.element != None:
                return self.element
            return False
        except:
            return False

class wait_for_text_to_start_with(object):
    def __init__(self, element, text, ignore_case=False):
        self.ignore_case = ignore_case
        self.element = element
        self.text = text

    def __call__(self, driver):
        try:
            element_text = self.element.text
            if self.ignore_case:
                element_text = element_text.lower()
                self.text = self.text.lower()
            return element_text.startswith(self.text)
        except:
            return False

class wait_for_text_to_end_with(object):
    def __init__(self, element, text, ignore_case=False):
        self.ignore_case = ignore_case
        self.element = element
        self.text = text

    def __call__(self, driver):
        try:
            element_text = self.element.text
            if self.ignore_case:
                element_text = element_text.lower()
                self.text = self.text.lower()
            return element_text.endswith(self.text)
        except StaleElementReferenceException:
            self.element.reload()
        except:
            return False

class wait_for_text_to_contain(object):
    def __init__(self, element, text, ignore_case=False):
        self.ignore_case = ignore_case
        self.element = element
        self.text = text

    def __call__(self, driver):
        try:
            element_text = self.element.text
            if self.ignore_case:
                element_text = element_text.lower()
                self.text = self.text.lower()
            if self.text in self.element.text:
                return True
            else:
                return False
        except:
            return False

class wait_for_attribute_value(object):
    def __init__(self, element, attribute, value):
        self.element = element
        self.attribute = attribute
        self.value = value

    def __call__(self, driver):
        try:
            if self.element.get_attribute(self.attribute) == self.value:
                return True
            else:
                return False
        except StaleElementReferenceException:
            self.element.reload()
            return False
        except:
            return False


class wait_for_css_attribute_value(object):
    def __init__(self, element, attribute, value):
        self.element = element
        self.attribute = attribute
        self.value = value

    def __call__(self, driver):
        try:
            if self.element.value_of_css_property(self.attribute) == self.value:
                return self.element
            else:
                return False
        except:
            return False




import unittest
import sys
import os
import builtins
import threading
from appium import webdriver
from appium_selector.DeviceSelector import DeviceSelector
from hotdog import Mustard
from hotdog.Results import UploadResults
from hotdog.Config import GetConfig
from hotdog.BaseDriver import BaseWebDriver
from hotdog.TestStep import StepLog, Step
from sauceclient import SauceClient
from selenium import webdriver as seleniumWebdriver
from selenium.webdriver import DesiredCapabilities
from time import sleep, time


class HotDogBaseTest(unittest.TestCase):

    # HotDog Defaults.  Can be overridden in child classes
    DefaultWebDriver = BaseWebDriver
    failed = False
    defaultTestResult = UploadResults

    #Boilerplate Settings Do not Change
    #Change in Config.xml
    SAUCE_USERNAME = GetConfig('SAUCE_USERNAME')
    SAUCE_ACCESS = GetConfig('SAUCE_ACCESS')

    SAUCE_URL = "http://%s:%s@ondemand.saucelabs.com:80/wd/hub" % (SAUCE_USERNAME, SAUCE_ACCESS)
    GRID_URL = GetConfig('GRID_URL') + '/wd/hub'
    LOCAL_APPIUM_URL = GetConfig('LOCAL_APPIUM_URL')

    # set SKIP_SELECTOR to bypass the device selector & run local with platfrom from LOCAL_BROWSER
    try:
        SKIP_SELECTOR = GetConfig('SKIP_SELECTOR')
        LOCAL_BROWSER = GetConfig('LOCAL_BROWSER')
        DEFAULT_CONFIG = {
                    'desiredCaps': {'browserName':  'Local',
                                    'deviceName':   'Local',
                                    'platformName': 'Local',
                                    'version':      'Local'},
                    'options': {'manufacturer': 'local',
                                'mustard':       False,
                                'provider':     'local-'+ LOCAL_BROWSER,
                                'osv':          'Local',
                                'model':        'local',
                                "idleTimeout":   360
                                }
                }
    except:
        #xml file is missing SKIP_SELECTOR/LOCAL_BROWSER
        SKIP_SELECTOR = 'False'


    timersStart = {}
    timersTotal = {}
    testcase_id = ''
    try:
        app_env =  os.environ['APP_DATA'].split('<|>')
        app_env = app_env[0]
        print('APP_DATA Environment Data found.  Appending app_env [%s] to test' % app_env[0])
        app_url = app_env[1]
        print('APP_DATA Environment Data found.  Appending app_url [%s] to test' % app_env[1])
    except:
        pass



    @classmethod
    def setUpClass(cls, platform='mobile'):
        if not hasattr(builtins, 'threadlocal'):

            builtins.threadlocal = threading.local()

            #Use Default Driver or display device selector
            if HotDogBaseTest.SKIP_SELECTOR == 'True':
                builtins.threadlocal.config = HotDogBaseTest.DEFAULT_CONFIG
            else:
                builtins.threadlocal.config = DeviceSelector(platform=platform).getDevice()[0]

            # If testing mobile apps, append app info into desired caps
            if 'platformName' in builtins.threadlocal.config['desiredCaps']:
                if builtins.threadlocal.config['desiredCaps']['platformName'].upper() == 'ANDROID':
                    builtins.threadlocal.config['desiredCaps']['app'] = GetConfig('ANDROID_APP_URL')
                else:
                    builtins.threadlocal.config['desiredCaps']['app'] = GetConfig('IOS_APP_URL')

            # Instantiate Driver
            desired_caps = builtins.threadlocal.config['desiredCaps']
            provider = builtins.threadlocal.config['options']['provider']
            try:
                builtins.threadlocal.driver = cls.select_driver(cls, desired_caps, provider)
            except:
                print(sys.exc_info()[1])
                raise unittest.SkipTest('Could not launch driver')


    def setUp(self):

        # Setup Test
        self.desired_caps = builtins.threadlocal.config['desiredCaps']
        self.options = builtins.threadlocal.config['options']
        self.options['deviceName'] = self.environmentName()
        self.provider = self.options['provider']
        self.resultLink = None

        # Reinitialize driver if destroyed
        if not builtins.threadlocal.driver:

            try:
                self.driver = self.__class__.select_driver(self, self.desired_caps, self.provider)
                builtins.threadlocal.driver = self.driver
                self.continueWithDriver = False
            except:
                print("Testcase [%s] COULD NOT START on device [%s]" % (self._testMethodName, self.options['deviceName']))
                print(sys.exc_info()[1])
                raise unittest.SkipTest('Could not launch driver')

        else:
            self.driver = builtins.threadlocal.driver
            self.continueWithDriver = True

        # Create root node for Step Log
        builtins.threadlocal.driver.step_log = StepLog()
        self.end_step = self.add_test_step(self._testMethodName)

        print("Testcase [%s] started on device [%s]" % (self._testMethodName, self.options['deviceName']))


    def environmentName(self):
        if 'deviceName' in self.options:
            return self.options['deviceName']
        elif 'udid' in self.desired_caps:
            return self.desired_caps['udid']
        else:
            return self.options['provider']

    def tearDown(self):
        if 'sauce' in self.provider.lower():
            sauce = SauceClient(self.SAUCE_USERNAME, self.SAUCE_ACCESS)
            self.resultLink = "https://saucelabs.com/beta/tests/%s" % self.driver.session_id
            try:
                if sys.exc_info() == (None, None, None):
                    sauce.jobs.update_job(self.driver.session_id, passed=True, name=self._testMethodName)
                else:
                    sauce.jobs.update_job(self.driver.session_id, passed=False, name=self._testMethodName)
            except:
                pass


    @classmethod
    def tearDownClass(cls):
        cls.RemoveApp()

    @classmethod
    def RemoveApp(self):
        if 'mobile' in builtins.threadlocal.config['options']['provider']:
            try: builtins.threadlocal.driver.close_app()
            except: pass
            try: builtins.threadlocal.driver.remove_app(GetConfig('IOS_BUNDLE_ID'))
            except: pass
            try: builtins.threadlocal.driver.remove_app(GetConfig('ANDROID_BUNDLE_ID'))
            except: pass
        try: builtins.threadlocal.driver.quit()
        except: pass

    def add_test_step(self, step_name):
        self.driver.step_log.add_step(Step(step_name))
        return self.driver.step_log.close_step

    def run(self, result=None):
        try:
            super().run( result= self.defaultTestResult())
        except:
            try:
                self.driver.quit()
            except:
                pass
            raise

    def timerStart(self, name):
        self.timersStart[name] = time()

    def timerStop(self, name):
        self.timersTotal[name] = time() - self.timersStart[name]
        if self.options['mustard']:
            Mustard.UploadPerformance(Mustard.getDeviceID(self), name, self.timersTotal[name])

    @staticmethod
    def select_driver(cls, dc, provider):
        runLocal = False
        browser_profile = None

        if 'grid' in provider:
            # if self.desired_caps['browserName'].lower() == 'firefox':
            #     browser_profile = seleniumWebdriver.FirefoxProfile('/home/matt/l7539ezh.GridTest')
            if dc['platform'].lower() == 'windows':
                dc['marionette'] = False
            url = cls.GRID_URL
        elif 'sauce' in provider:
            url = cls.SAUCE_URL
            if dc['browserName'] == 'internet explorer':
                dc['requireWindowFocus'] = True
            elif dc['browserName'] == 'chrome' and dc.get('chromeOptions') is None:
                builtins.threadlocal.config['desiredCaps']['chromeOptions'] = {
                    'excludeSwitches': ['disable-component-update']}
        elif provider.lower() == 'local-chrome':
            runLocal = True
            driver = seleniumWebdriver.Chrome()
        elif provider.lower() == 'local-firefox:marionette':
            runLocal = True
            caps = DesiredCapabilities.FIREFOX
            caps['marionette'] = True
            driver = seleniumWebdriver.Firefox(caps)
        elif provider.lower() == 'local-firefox':
            caps = DesiredCapabilities.FIREFOX
            caps['marionette'] = False
            runLocal = True
            driver = seleniumWebdriver.Firefox(caps)
        elif provider.lower() == 'local-safari':
            runLocal = True
            driver = seleniumWebdriver.Safari()
        elif provider.lower() == 'local-ie':
            runLocal = True
            driver = seleniumWebdriver.Ie()
        elif provider.lower() == 'mcweb':
            url = GetConfig('MC_URL') + '/wd/hub'
        elif provider.lower() == 'mcmobile':
            url = GetConfig('MC_URL') + '/wd/hub'
        else:
            url = cls.GRID_URL

        if not runLocal:
            driver = webdriver.Remote(
                url,
                dc,
                browser_profile=browser_profile
            )

        driver.__class__ = cls.DefaultWebDriver
        driver.test_steps = []
        driver.tracelog = []

        return driver




import builtins
import json
import time
import re


def TestStep(argument=False, log_result=False):
    def test_step(function):
        def wrapper(*args, **kwargs):

            step = Step(argument)

            matches = re.findall('\{(.*?)\}', step.step_name)
            for m in matches:
                try:
                    step.step_name = step.step_name.replace('{%s}' % m, str(eval(m)))
                except:
                    step.step_name = step.step_name.replace('{%s}' % m, '<ERROR>')

            builtins.threadlocal.driver.step_log.add_step(step)
            try:
                result = function(*args, **kwargs)
            except:
                step.end_step('error')
                raise
            else:
                if log_result:
                    result_step = Step('Result [%s]' % result)
                    builtins.threadlocal.driver.step_log.add_step(result_step)
                    result_step.end_step('complete')
                    builtins.threadlocal.driver.step_log.close_step()
                step.end_step('complete')
            builtins.threadlocal.driver.step_log.close_step()
            return result

        return wrapper
    return test_step


class Step(object):

    def __init__(self, step_name):
        super().__init__()
        self.step_name = step_name
        self.sub_steps = []
        self.start_time = time.time()
        self.end_time = ''
        self.elapsed_time = ''
        self.status = ''

    def add_step(self, sub):

        self.sub_steps.append(sub)

    def end_step(self, status):
        self.status = status
        self.end_time = time.time()
        self.elapsed_time = self.end_time - self.start_time


def encode_step(obj):
    if isinstance(obj, Step):
        return obj.__dict__
    else:
        return obj


class StepLog(object):

    def __init__(self):
       self.steps = []
       self.open_step = []

    def add_step(self, step):
        if self.open_step:
            self.open_step[-1].add_step(step)
            self.open_step.append(step)
        else:
            self.steps.append(step)
            self.open_step.append(step)

    def close_step(self, status='complete'):
        self.open_step[-1].end_step(status)
        del self.open_step[-1]

    def to_json(self):
        return json.dumps(self.steps, default=encode_step)




import time


def Retry(func, timeout=30):
    def inner(*args, **kwargs):
        start = time.time()
        while True:
            try:
                ret = func(*args, **kwargs)
                return ret
            except:
                if time.time() - start > timeout:
                    raise
    return inner


import time

from hotdog.TestStep import Step, TestStep
from selenium.common.exceptions import WebDriverException, NoSuchElementException
from random import randint
from hotdog import Mustard
from hotdog.BaseDriver import get_driver
from waiting import wait

class HotDogBasePage(object):

    def __new__(cls, *args, **kwargs):
        if 'driver' in kwargs:
            driver = kwargs['driver']
            try:
                platform = driver.desired_capabilities['desired']['platformName'].lower()
            except:
                return super().__new__(cls)
            if hasattr(cls,platform):
                return getattr(cls, platform)(driver=driver)
            else:
                return super().__new__(cls)
        else:
            return super().__new__(cls)

    def __init__(self, driver=None, url=None):
        self.driver = driver
        self.url = url

    def sync(self, timeout=20):
        self.driver.implicitly_wait(timeout)
        if hasattr(self, 'sync_element'):
            self.find('sync_element')
        else:
            time.sleep(5)

    def find(self, objectName, type=None):
        locators = getattr(self, objectName)
        self.driver.implicitly_wait(0)
        if len(locators) == 3 and not type:
            type = locators[2]
        try:
            element = self.driver.find_element(locators[0], locators[1], type=type, name=objectName)
        except:
            if type:
                element = type(self.driver, None, by=locators[0], value=locators[1], name=objectName, type=type)
            else:
                element = self.driver.DefaultElementType(self.driver, None, by=locators[0], value=locators[1], name=objectName, type=type)
            element._driver = self.driver
            element._parent = self.driver
        self.driver.implicitly_wait(self.driver.default_wait_time)
        return element

    def finds(self, objectName, type=None):
        self.driver.implicitly_wait(0)
        locators = getattr(self, objectName)
        if len(locators) == 3 and not type:
            type = locators[2]
        element = self.driver.find_elements(locators[0], locators[1], type=type, name=objectName)
        self.driver.implicitly_wait(self.driver.default_wait_time)
        return element

    def find_random(self, object_name, type=None, timeout=20):
        '''Returns a random collection element.
        :param object_name: the object name
        :return: randomly selected element
        '''
        locators = getattr(self, object_name)
        if len(locators) == 3 and not type:
            type = locators[2]
        elements = self.driver.find_elements(locators[0], locators[1], type=type)
        element_count = elements.count(timeout=timeout)
        if element_count == 0:
            raise NoSuchElementException('Element Not Found [%s]' % elements)
        index = randint(0, len(elements) - 1)
        return elements[index]

    def find_element(self, *args, **kwargs):
        self.driver.implicitly_wait(0)
        try:
            element =  self.driver.find_element(*args, **kwargs)
            element._parent = self.driver
        except:
            element =  self.driver.DefaultElementType(self.driver, None, by=kwargs['by'], value=kwargs['value'])
            element._driver = self.driver
            element._parent = self.driver
        self.driver.implicitly_wait(self.driver.default_wait_time)
        return element

    def find_elements(self, *args, **kwargs):
        elements =  self.driver.find_elements(*args, **kwargs)
        elements._driver = self.driver
        elements._parent = self.driver
        return elements


    def back(self):
        self.driver.execute_script("window.history.go(-1)");

    def swipe(self, direction, element=None, duration=None):
        if element:
            location = element.location
            size = element.size
            center_x_loc = location['x'] + (size['width']/2)
            center_y_loc = location['y'] + (size['height']/2)
        else:
            size = self.driver.get_window_size()
            center_x_loc = size['width']/2
            center_y_loc = size['height']/2

        screen = self.driver.get_window_size()

        if direction.lower() == 'left':
                start_x = screen['width'] * 0.9
                end_x = screen['width'] * 0.1
                start_y = center_y_loc
                end_y = center_y_loc

        elif direction.lower() == 'right':
                start_x = screen['width'] * 0.01
                end_x = screen['width'] * 0.9
                start_y = center_y_loc
                end_y = center_y_loc

        elif direction.lower() == 'up':
                start_x = center_x_loc
                end_x = center_x_loc
                start_y = screen['height'] * 0.9
                end_y = screen['height'] * 0.1

        elif direction.lower() == 'down':
                start_x = center_x_loc
                end_x = center_x_loc
                start_y = screen['height'] * 0.1
                end_y = screen['height'] * 0.9
        else:
            raise ValueError('Invalid Direction for swipe. [%s]' % direction)

        self.driver.swipe(start_x, start_y, end_x, end_y, duration=duration)

    def open(self):
        if not self.url:
            raise Exception('Can\'t open page without url')
        self.driver.get(self.url)

    def implicitly_wait(self, *args):
        return self.driver.implicitly_wait(*args)

    def is_element_present(self, element_name, just_in_dom=False, timeout=0):
        def _get_driver():
            driver = getattr(self, 'driver', None)
            if driver:
                return driver
            return get_driver()

        _get_driver().implicitly_wait(timeout)
        try:
            def is_displayed():
                element = getattr(self, element_name, None)
                if not element:
                    raise Exception('No element "%s" within container %s' % (element_name, self))
                return element.is_displayed()

            is_displayed() if just_in_dom else self.wait(lambda: is_displayed(), timeout_seconds=timeout)
            return True
        except Exception:
            return False
        except TimeoutError:
            return False

    def assert_element_present(self, elementName, timeout=10):
        assert self.is_element_present(elementName, timeout=timeout), 'The element [%s] was not found after [%s] seconds' % (elementName, timeout)

    def wait(self, *args, **kwargs):
        """
        Wrapping 'wait()' method of 'waiting' library with default parameter values.
        WebDriverException is ignored in the expected exceptions by default.
        """
        kwargs.setdefault('sleep_seconds', (1, None))
        kwargs.setdefault('expected_exceptions', WebDriverException)
        kwargs.setdefault('timeout_seconds', 30)

        return wait(*args, **kwargs)

    def uploadScreenshot(self, test, name=None):
        Mustard.UploadScreenshot(self, test, name);

    def add_test_step(self, step_name):
        self.driver.step_log.add_step(Step(step_name))
        return self.driver.step_log.close_step

    @TestStep('Click on {args[3]} at coordinates: x - {args[1]}, y - {args[2]}')
    def click_at(self, x, y, message='location'):
        self.driver.execute_script("$(document.elementFromPoint(" + str(x) + "," + str(y) + ")).click();")


from selenium.webdriver.common.by import By
from appium.webdriver.webdriver import WebDriver
from hotdog.BaseElements import BaseElements
from hotdog.BaseElement import BaseElement


class BaseWebDriver(WebDriver):

    DefaultElementType = BaseElement
    default_wait_time = 5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implicitly_wait(self.default_wait_time)

    @property
    def driver(self):
        return self

    def implicitly_wait(self, time_to_wait):
        pass

    def find_element(self, by=By.ID, value=None, type=None, name=None):
        self.implicitly_wait(0)
        try:
            element =  super().find_element(by, value)
        except:
            if type:
                element = type(self.driver, None, by=by, value=value, name=name, type=type)
            else:
                element = self.driver.DefaultElementType(self.driver, None, by=by, value=value, name=name, type=type)
            element._parent = self.driver

        if type:
            element.__class__ = type

        element.search_by = (by, value, type)
        element.by = by
        element.value = value
        if name:
            element.name = name

        if hasattr(self, 'debug'):
            element.debug = self.debug
        self.implicitly_wait(self.default_wait_time)
        return element

    def find_elements(self, by=By.ID, value=None, type=None, name=None):

        elements =  super().find_elements(by, value)

        for element in elements:
            if type:
                element.__class__ = type
            if hasattr(self, 'debug'):
                element.debug = self.debug

        if len(elements) > 0:
            loaded = True
        else:
            loaded = False

        #Wrap Element Collection with BaseElements
        return BaseElements(elements, self, by, value, type=type, name=name, loaded=loaded)

    def create_web_element(self, element_id):
        return self.DefaultElementType(self, element_id, w3c=self.w3c)

    def get_new_handle(self, existing_handles):
    # IE doesn't guarantee the order of window handles when a new window is opened
    #
    # @param existing_handles The handles before the action of opening a new window
        for handle in super().window_handles:
            if handle not in existing_handles:
                return handle

def get_driver():
    global _driver_instance
    if not _driver_instance:
        _driver_instance = BaseWebDriver

    return _driver_instance